# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import re
import time
import gtk
import pango
import gobject

from emesenecommon import unescape
from emesenecommon import escape
from emesenecommon import rgbToHexa
from emesenecommon import getLinks
from msnPlusMarkup import msnPlusMarkup

import ConversationUI
import Log

from cgi import escape

tagStyles = {
    'nudge':'font-weight:bold; color: black',
    'donudge':'font-weight: bold; color: black',
    'me':'font-weight: bold; color: blue',
    'tstamp':'font-weight: normal',
    'sender':'font-weight: bold; color: red',
    'message': 'color: black'
    }

class Conversation( object ):
    '''This class is an abstraction of a conversation, it is used to separate the
    data from the GUI in a conversation to let us have single and tabbed windows
    with the same codebase (also MVC is good :P)'''

    def __init__( self, switchboard, theme, config, textTagTable ):
        '''Contructor'''

        self.setSwitchboard( switchboard )
        self.config = config
        self.textTagTable = textTagTable
        
        self.contactManager = switchboard.msn.contactManager
        self.lastMessageMail = '' # the mail of the user who sent the last message
        self.isCurrent = False # if True this is the tab that has the focus
        self.tabNum = None # the tabnumber of this conversation
        self.closed = False
        self.textBuffer = gtk.TextBuffer( self.textTagTable )
        self.autoreplySent = False # if true we allready sent the autoreply message
        self.log = False
        
        self.inputText = ''
        self.theme = theme
        self.window = None
        
        switchboard.msn.connect( 'personal message changed', self.onPersonalMessageChanged )
        switchboard.msn.connect( 'nick changed', self.onNickChanged )
        switchboard.msn.connect( 'contact status change', self.onStatusChanged )
        switchboard.msn.connect( 'color changed', self.onColorChanged )
        switchboard.msn.connect( 'font changed', self.onFontChanged )
        
        self.ui = ConversationUI.ConversationUI( self, self.config, self.theme )
        self.color = self.config.getValue( 'color', '#000000' ).replace( '#', '' )
        self.font = self.config.getValue( 'font' ).replace( '%20', ' ' )

        self.bold = False
        self.italic = False
        self.underline = False
        self.setSize(10*pango.SCALE)
        
        # here we put the conversation, if we have tabbed windows we switch the
        # textbuffer according to the tab

        path = self.config.getLogPath()
        filename = self.switchboard.getUser() + time.strftime( "%d-%m-%Y %H%M%S" )
        self.log = Log.Log( self.theme, path, filename )

        self.switchboard.setTypingCallback(self.receiveTyping)
        
        self.plusMarkup = msnPlusMarkup()
        
    def close( self ):
        '''close the tab'''
        
        self.window.close_tab( self, self.tabNum )
        
    def onColorChanged( self, colorStr ):
        self.setColor(colorStr)
        self.ui.getInput().applyAttrsToInput()
    
    def onFontChanged( self, font, bold, italic, size ):
        self.setFont(font, italic, bold, size)
        
        self.ui.toolbar.setBold( bold )
        self.ui.toolbar.setItalic( italic )
        
        self.ui.getInput().applyAttrsToInput()
        
    def onPersonalMessageChanged( self, email, personalMessage ):
        if email in self.getMembers():
            self.ui.update()
        
    def onNickChanged(self, email, nick ):
        if email in self.getMembers():
            self.ui.update()
            
            if self.isCurrent:
                self.window.set_title( self.getTitle() )
        
    def onStatusChanged( self, email, status ):
        if email in self.getMembers():
            self.ui.update()
        
    def getUseTimestamp( self ):
        '''return the timestamp value'''
        return self.config.getValue( 'showTimestamp' ) == '1'

    def addToLog( self, text, style  ):
        '''add text to the log'''

        self.log.add( text, style )

    def saveLog( self ):
        '''save the log to a file.'''
        self.log.save()

    def setColor( self, color ):
        '''set the color of the user text'''
        self.color = color.replace( '#', '' )

    def getColor( self ):
        '''return the user color'''
        return self.color

    def setFont( self, font, italic=False,
                 bold=False, size=10 ):
        '''set the font of the user text'''
        self.font = font
        self.italic = italic
        self.bold = bold
        self.size = size

    def getFont( self ):
        '''return the user font'''

        return self.font

    def setBold( self, value ):
        '''set the attribute bold to value'''

        self.bold = value

    def getBold( self ):
        '''return the bold value'''

        return self.bold
    
    def setSize( self, value ):
        '''set the attribute size to value'''

        self.size = value

    def getSize( self ):
        '''return the size value'''

        return self.size
    
    def setItalic( self, value ):
        '''set the attribute italic to value'''

        self.italic = value

    def getItalic( self ):
        '''return the log value'''
        return self.italic

    def setUnderline( self, value ):
        '''set the attribute underline to value'''

        self.underline = value

    def getUnderline( self ):
        '''return the underline value'''

        return self.underline

    def getTextBuffer( self ):
        '''return the textbuffer'''

        return self.textBuffer

    def getTitle( self ):
        '''return a short title according to the users in the conversation'''

        MAX_TITLE_LENGHT = 30

        if len( self.switchboard.getMembers() ) > 1:
            return _( 'Group chat' )
        elif len( self.switchboard.getMembers() ) == 1:
            # switchboard.getNick gives escaped string
            nick = unescape( self.switchboard.getNick( self.switchboard.getMembers()[ 0 ] ) )
            if len( nick ) > MAX_TITLE_LENGHT:
                return  nick[:MAX_TITLE_LENGHT - 3] + '...'
            else:
                return nick
        elif len( self.switchboard.getInvitedMembers() ) >= 1:
            # switchboard.getNick gives escaped string
            title = unescape( self.switchboard.getNick( self.switchboard.getInvitedMembers()[ 0 ] ) )
            if len( title ) > MAX_TITLE_LENGHT:
                return  title[:MAX_TITLE_LENGHT - 3] + '...'
            else:
                return title
        else:
            return ''

    def getUser( self ):
        '''return the user mail'''

        return self.switchboard.getUser()

    def getStyle( self ):
        '''return the style string to use in the sendMessage method'''

        effectValue = ''

        if self.bold:
            effectValue += 'B'

        if self.italic:
            effectValue += 'I'

        if self.underline:
            effectValue += 'U'

        color = self.color[ 4:6 ] + self.color[ 2:4 ] + self.color[ :2 ]

        return "X-MMS-IM-Format: FN=" + self.font.replace( ' ', '%20' ) + "; EF=" + effectValue + "; CO=" + color + "; PF=0"

    def getCssStyle( self ):
        '''return the css style of the user'''
        
        effectValue = ''

        if self.bold:
            effectValue += 'font-weight: bold;'

        if self.italic:
            effectValue += 'font-style: italic;'

        if self.underline:
            effectValue += 'text-decoration: underline;'

        effectValue += 'color: #' + self.color + ';'
        effectValue += 'font-family: ' + self.font + ';'
        
        return effectValue

    def getId( self ):
        '''return the id of the switchboard'''

        return self.switchboard.getId()

    def getOnlineUsers( self ):
        '''This method returns a list ol mails of the contacts who are not offline'''
        return self.switchboard.getOnlineUsers()

    def invite( self, mail ):
        '''invite a user to the conversation'''

        self.switchboard.invite( mail )

    def getMembers( self ):
        '''return a list of the members in the conversation'''
        return self.switchboard.getMembers()

    def getWindow( self ):
        '''return the window that hold this conversation'''
        return self.window

    def setWindow( self, window ):
        '''set the window that hold this conversation'''

        self.window = window

    def setIsCurrent(self, current ):
        '''set the isCurrent attribute, if true, this conversation
        is the tab that is shown'''
        self.isCurrent = current

    def getIsCurrent(self):
        '''return the value of isCurrent'''
        return self.isCurrent

    def receiveNudge( self, mail ):
        '''This method is called when a nudge is received in the
        switchboard'''
        
        self.appendOutputText( None, _( "%s just sent you a nudge!\n" )%self.switchboard.getUserDisplayName( mail ), tagStyles[ 'nudge' ] )
        self.window.setUrgency()

    def receiveMessage( self, mail, nick, message, format, charset ):
        '''This method is called when a message is received in the
        switchboard'''

        if not self.window.has_toplevel_focus() or not self.isCurrent:
            self.switchboard.msn.emit( 'message received' )
            self.ui.setMessageWaiting()

        if self.config.getValue( 'autoReply' ) == '1' and not self.autoreplySent:
            self.switchboard.sendMessage( 'AutoMessage: ' + self.config.getValue( 'autoReplyMessage' ).replace( '%20', ' ' ) )
            self.appendOutputText( None, 'AutoMessage: %s\n'%self.config.getValue( 'autoReplyMessage' ).replace( '%20', ' ' ), tagStyles[ 'nudge' ] )
            self.autoreplySent = True

        if self.getUseTimestamp():
            self.addToLog( self.getTimeStamp() + ' ', tagStyles[ 'tstamp' ] )

        self.appendOutputText( mail , message + '\n', self.parseFormat( mail, format ), tagStyles[ 'sender' ] )

        self.window.setUrgency()
        self.window.show()

    def userJoin( self, mail ):
        '''This method is called when someone joing the conversation'''
        
        if self.isCurrent:
            self.window.set_title( self.getTitle() )

        self.ui.update()

    def userLeave( self, mail ):
        '''method called when someone leaves the conversation'''
        self.appendOutputText( None, _( "%s has left the conversation" )%self.switchboard.getUserDisplayName( mail ) + '\n', tagStyles[ 'nudge' ] )
        
        self.ui.update()
        
        if self.isCurrent:
            self.window.set_title( self.getTitle() )
        
    def userOffline( self, mail ):
        '''method called when someone goes offline in the conversation'''
        self.appendOutputText( None, _( "%s is now offline" )%self.switchboard.getUserDisplayName( mail ) + '\n', tagStyles[ 'nudge' ] )
        
        self.ui.update()

    def userOnline( self, mail ):
        '''method called when someone goes online in the conversation'''
        self.appendOutputText( None, _( "%s is now online" )%self.switchboard.getUserDisplayName( mail ) + '\n', tagStyles[ 'nudge' ] )

    def inviteUser( self, mail ):
        '''method called when the user select a frined in the invite dialog'''
        self.invite( mail )
        
        self.ui.update()

    def doNudge( self ):
        '''this method is called when the user click the nudge button'''
        self.switchboard.sendNudge()
        self.appendOutputText( None, _( "you have sent a nudge!" ) + '\n', tagStyles[ 'donudge' ] )

    def reconnect( self ):
        '''reconnect the switchboard'''

        users = self.switchboard.getMembers()
        msn = self.switchboard.msn
        self.setSwitchboard( msn.newSwitchboard() )

        for i in users:
            self.switchboard.invite( i )

        self.autoreplySent = False

    def sendMessage( self, message ):
        '''send a message to the conversation'''
        if message[0]=='/':
            self.switchboard.msn.emit( 'slash command', self, message )
            return
            
        if self.switchboard._status == 'closed':
            self.reconnect()

        try:
            self.switchboard.sendMessage( message, self.getStyle() )
        except Exception, e:
            print str( e )
            self.reconnect()
            self.sendMessage(message)
            return

        self.appendOutputText( self.window.user, message + '\n', self.getCssStyle(), tagStyles[ 'me' ] )
            

    def sendIsTyping( self ):
        '''a easy method to send the is typing message'''
        if self.switchboard._status == 'closed':
            self.reconnect()

        try:
            self.switchboard.sendIsTyping()
        except Exception:
            self.reconnect()
            self.sendIsTyping()

    def parseFormat( self, mail, format ):
        '''parse the format of a mail and return the style'''

        # FN=Sans; EF=; CO=000000; PF=0
     
        style = ''
     
        if format.find( "FN=" ) != -1:
            font = format.split( 'FN=' )[ 1 ].split( ';' )[ 0 ].replace( '%20', ' ' )
            style += 'font-family: ' + font + ';'

        if format.find( "CO=" ) != -1:
            color = format.split( 'CO=' )[ 1 ].split( ';' )[ 0 ]
            if len( color ) == 6:
                color = color[ 4:6 ] + color[ 2:4 ] + color[ :2 ]
                style += 'color: #' + color + ';'
            elif len( color ) == 3:
                color = color[ 2 ] + color[ 1 ] + color[ 0 ]
                style += 'color: #' + color + ';'

        if format.find( "EF=" ) != -1:
            effect = format.split( 'EF=' )[ 1 ].split( ';' )[ 0 ]

            if effect.find( "B" ) != -1:
                style += 'font-weight: bold;'

            if effect.find( "I" ) != -1:
                style += 'font-style: italic;'

            if effect.find( "U" ) != -1:
                style += 'text-decoration: underline;'

        return style

    def getTimeStamp( self ):
        '''return a string representating the current time'''

        return '(' + time.strftime( '%H:%M:%S' ) + ')'

    def setSwitchboard( self, switchboard ):
        '''
        set a new sitchboard for the conversation.
        usefull if the conversation is closed and the other user start a new one
        '''

        self.switchboard = switchboard

        self.switchboard.setNudgeCallback( self.receiveNudge )
        self.switchboard.setMessageCallback( self.receiveMessage )
        self.switchboard.setUserJoinCallback( self.userJoin )
        self.switchboard.setUserLeaveCallback( self.userLeave )
        self.switchboard.setUserOfflineCallback( self.userOffline )
        self.switchboard.setUserOnlineCallback( self.userOnline )

        self.autoreplySent = False

    def appendOutputText( self, username, text, style = None, nickStyle = None):
        '''append the given text to the outputBuffer'''
        smileys = {}
        text = escape(text)
        text = getLinks( text )
        
        if style:
            text = '<span style="'+style+'">' + text + '</span>'
        
        if username != None:
            if self.getUseTimestamp():
                timestamp = self.getTimeStamp() + ' '
            else:
                timestamp = ''
            if nickStyle:
                if username == self.switchboard.getUser():
                    nick = self.switchboard.msn.getNick()
                    formattedNick = '<span style="'+nickStyle+'">' + timestamp + escape(nick) + ': </span>'
                else:
                    formattedNick = '<span style="'+nickStyle+'">' + timestamp + escape(self.switchboard.getUserDisplayName( username )) + ': </span>'
            else:
                formattedNick = timestamp + escape(self.switchboard.getUserDisplayName( username )) + ': '
            
            text = text.replace("\n", '<br/>')
            formattedNick = self.plusMarkup.replaceMarkup( '<span>'+formattedNick+'</span>', True )
            text = self.plusMarkup.replaceMarkup( '<span>'+text+'</span>', True )
            text = formattedNick + text

        for i in self.process( [ text ] , self.theme.getSmileysList() ):
            smiley = self.theme.getSmileyPath( i )
            if smiley:
                smileys[i] = '<img src="file://'+smiley+'"/>'
                
        for key, value in smileys.iteritems():
            text = text.replace(key, value)

        self.addToLog(text, '')
        try:
            self.ui.textview.display_html( '<span>'+text+'</span>'.encode() )
        except Exception, e:
            print 'error trying to display "' + text + "'" 
            print e

        self.window.scrollToBottom( self.tabNum )
                
        self.textBuffer.place_cursor( self.textBuffer.get_end_iter() )
        self.window.scrollToBottom( self.tabNum )

    def process( self , list , smilies ):
        '''create a list of text and smilies like [ "hi, mi name is ", ":)" , " pepe" ]'''

        if smilies == []:
            return list

        newlist = []

        for i in list:
            for j in i.split( smilies[ 0 ] ):
                newlist += [ j ]
                newlist += [ smilies[ 0 ] ]

            newlist = newlist[ :-1 ]

        return self.process( newlist , smilies[ 1: ] )

    def getStatus( self ):
        return self.switchboard.getStatus()

    def setStatus( self, value ):
        self.switchboard.setStatus( value )

    def isClosed( self ):
        return self.closed

    def setClosed( self, value ):
        self.closed = value
        self.switchboard.leaveChat()
    
    def receiveTyping( self, mail ):
        '''This method is called when a is typing message is received in the
        switchboard'''

        self.ui.setTyping()

    def getMembersDict( self ):
        '''return a dict with email as key and contact instance as value'''
        
        userDict = {}
        for i in self.getMembers():
            if self.contactManager.getContact( i ) != None:
                userDict[ i ] = self.contactManager.getContact( i )
            else:
                userDict[ i ] = self.contactManager.getDummyContact( i )
        
        return userDict
    def getTextTagTable( self ):
        '''return the texttagtable'''
        return self.textTagTable

    def getTextTag( self ):
        '''return a text tag from the current style'''
        
        tag = gtk.TextTag()
        
        tag.set_property( 'font', self.font )
        tag.set_property( 'size', self.size )
        tag.set_property( "foreground-gdk" , gtk.gdk.color_parse( "#" + self.color ) )
        
        if self.bold:
            tag.set_property( "weight" , pango.WEIGHT_BOLD )
        
        if self.italic:
            tag.set_property( "style" , pango.STYLE_ITALIC )
        
        if self.underline:
            tag.set_property( "underline-set" , True )
            tag.set_property( "underline" , pango.UNDERLINE_LOW )
        else:
            tag.set_property( "underline-set" , True )
            tag.set_property( "underline" , pango.UNDERLINE_NONE )
            
        return tag