# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import msnp13
from xml.dom import minidom
import emesenecommon as ec

class Contact( object ):
    '''class describing a contact'''

    def __init__( self, email, id = '', nick = '', personalMessage = '', alias = '',\
         status = 'FLN', mobile = False, blocked=False, space=False, allow=False,\
             reverse=False, pending= False, groups=None ):
        '''Contructor'''

        self.email = email.lower()
        self.id = id
        self.nick = nick
        self.personalMessage = personalMessage
        self.alias = alias
        self.status = status
        self.mobile = mobile
        self.space = space
        if groups:
            self.groups = groups # for performance
        else:
            self.groups = []

        self.allow = allow
        self.blocked = blocked
        self.reverse = reverse
        self.pending= pending
        
        self.msnobj = None

    def __repr__( self ):
        return self.email + ': ' + self.id + ' ' + self.nick + '\n'\
                         + str( self.groups )

    def getEmail( self ):
        return self.email

    def setEmail( self, value ):
        self.email = value.lower()

    def getId( self ):
        return self.id

    def setId( self, value ):
        self.id = value

    def getNick( self ):
        return self.nick

    def setNick( self, value ):
        self.nick = value

    def getPersonalMessage( self ):
        return self.personalMessage

    def setPersonalMessage( self, value ):
        self.personalMessage = value

    def getAlias( self ):
        return self.alias

    def setAlias( self, value ):
        self.alias = value

    def getStatus( self ):
        return self.status

    def setStatus( self, value ):
        self.status = value

    def getMobile( self ):
        return self.mobile

    def setMobile( self, value ):
        self.mobile = value

    def getAllow( self ):
        return self.allow

    def setAllow( self, value ):
        self.allow = value

    def getBlocked( self ):
        return self.blocked

    def setBlocked( self, value ):
        self.blocked = value

    def getPending( self ):
        return self.pending

    def setPending( self, value ):
        self.pending = value

    def getReverse( self ):
        return self.reverse

    def setReverse( self, value ):
        self.reverse = value

    def getSpace( self ):
        return self.space

    def setSpace( self, value ):
        self.space = value

    def getGroups( self ):
        return self.groups

    def setGroups( self, value ):
        self.groups = value

    def addGroup( self, id ):
        if not id in self.groups:
            self.groups.append( id )

    def removeGroup( self, id ):
        if id in self.groups:
            self.groups.remove( id )
        else:
            raise GroupNotFoundError, id

    def setMsnobj( self, msnobj ):
        self.msnobj = msnobj

    def getMsnobj( self ):
        return self.msnobj

class Group( object ):
    '''class representing a group'''

    def __init__( self, name, id = '' ):
        '''Contructor,
        users is a dict with an email as key and a contact object as value
        '''

        self.name = name
        self.id = id
        self.users = {}

    def getName( self ):
        return self.name

    def setName( self, value ):
        self.name = value

    def getId( self ):
        return self.id

    def setId( self, value ):
        self.id = value

    def getUsersByStatus(self, status):
        """Returns a list user users according to its status"""
        return [i for i in self.users.values() if \
             i.getStatus() == msnp13.status_table[status]]

    def getUsers( self ):
        '''return the dict'''
        return self.users

    def setUsers( self, value ):
        '''
        Sets the users of the group to the dictionary { email: contact }
        deleting any previously introduced contact
        '''
        self.users = value

    def getUser( self, email ):
        if self.users.has_key( email ):
            return self.users[ email ]
        else:
            return None

    def setUser( self, email, contactObject ):
        self.users[ email ] = contactObject

    def removeUser( self, email ):
        # if self.users.has_key( email ):
            del self.users[ email ]

    def getSize(self):
        """returns how many users the group has"""
        return len(self.users)

    def getOnlineUsersNumber( self ):
        '''
        Returns the number of online users (not offline) in the group
        '''
        
        offline = self.getUsersByStatus( 'offline' )
        return self.getSize() - len( offline )

class ContactList( object ):
    '''a class that contains groups that contains users'''

    def __init__( self, groups=None ):
        '''Constructor,
        groups is a dict with the name as key and a Group object as value'''

        self.groups = {}
        self.reverseGroups = {}
        self.noGroup = Group( 'No group', 'nogroup' )
        self.contacts = {}

        if groups:
            self.setGroups( groups )
            
        self.lists = {}
        self.lists[ 'Allow' ] = []
        self.lists[ 'Block' ] = []
        self.lists[ 'Reverse' ] = []
        self.lists[ 'Pending' ] = []

    def getGroups( self ):
        '''return the dict'''

        return self.groups
    
    def getGroupNames( self ):
        '''
        Returns a list with the groups' names in the contact list
        '''
        
        return self.reverseGroups.keys()

    def setGroups( self, groupDict ):
        '''set the dict'''

        self.groups = groupDict.copy()
        self.reverseGroups = {}
        for i in self.groups.keys():
            self.reverseGroups[ self.groups[ i ].getName() ] = self.groups[ i ]

    def addGroup( self, name, gid ):
        group = Group( name, gid )
        self.setGroup( gid, group )

    def getGroup( self, id ):
        if self.groups.has_key( id ):
            return self.groups[ id ]
        else:
            return None

    def setGroup( self, id, groupObject ):
        self.groups[ id ] = groupObject
        self.reverseGroups[ groupObject.getName() ] = self.groups[ id ]
        # update contained contacts
        for contact in groupObject.getUsers().copy():
            self.addUserToGroup( contact, id )

    def removeGroup( self, group ):
        if self.groups.has_key( group ):
            # update contained contacts
            for contact in self.groups[ group ].getUsers().copy():
                self.removeUserFromGroup( contact, group )
            name = self.groups[ group ].getName()
            # remove the group
            del self.groups[ group ]
            del self.reverseGroups[ name ]

    def renameGroup( self, id, newName ):
        if self.groups.has_key( id ):
            group = self.groups[ id ]
            self.removeGroup( id )
            group.setName( newName )
            self.setGroup( id, group )

    def getGroupId( self, name ):
        if self.reverseGroups.has_key( name ):
            return self.reverseGroups[ name ].getId()
        elif name == 'No group':
            return 'nogroup'
        else:
            return None
    def getGroupName( self, id):
        if id in self.groups:
            return self.groups[ id ].getName()
        elif id == 'nogroup':
            return 'No group'
        else:
            return None

    def addContact( self, contact ):

        self.contacts[ contact.getEmail() ] = contact

        for id in contact.getGroups():
            if self.groups.has_key( id ):
                self.groups[ id ].setUser( contact.getEmail(), contact )

        if contact.getGroups() == []:
            self.noGroup.setUser( contact.getEmail(), contact )
            
    def addNewContact( self, email, groups=None):        
        if email in self.lists[ 'Block' ]:
            contact = Contact( email, blocked=True )
        else:
            contact = Contact( email, allow=True )
        self.addContact( contact )
        '''id = '', nick = '', personalMessage = '', alias = '',\
         status = 'FLN', mobile = False, blocked=False, space=False, allow=False,\
             reverse=False, pending= False, groups=None )'''
             
    def setContactIdXml( self, email, xml ):
        '''
        Sets a contact's id from the add user soap response xml
        '''
        if email in self.contacts:
            ml = minidom.parseString( xml )
            response = ml.getElementsByTagName( 'ABContactAddResponse' )[ 0 ]
            guid = response.getElementsByTagName( 'guid' )[ 0 ].childNodes[ 0 ].data
            self.contacts[ email ].setId( guid )
        else:
            raise ContactNotInListError, email

    def getContacts( self ):
        return self.contacts

    def getContact( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ]
        else:
            return None
        
    def getDummyContact( self, email ):
        '''build a dummy contact with some data to be allowed to show data about
        contacts that we dont have i.e when someone add in a group chat someone that we dont have.'''
        
        return Contact(email,'',email,'','','NLN',False,False,False,True,True,False)
            
    def getContactStatus( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getStatus()
        else:
            raise ContactNotInListError, email

    def setContactStatus( self, email, status ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setStatus( status )
        else:
            raise ContactNotInListError, email

    def getContactHasSpace( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getSpace()
        else:
            raise ContactNotInListError, email

    def getContactHasMobile( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getMobile()
        else:
            raise ContactNotInListError, email

    def getContactIsBlocked( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getBlocked()
        else:
            raise ContactNotInListError, email

    def setContactIsBlocked( self, email, value ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setBlocked( value )
        else:
            raise ContactNotInListError, email

    def getContactIsAllowed( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getAllow()
        else:
            raise ContactNotInListError, email
        
    def setContactIsAllowed( self, email, value ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setAllow( value )
        else:
            raise ContactNotInListError, email

    def getContactNick( self, email, escaped=False ):
        if self.contacts.has_key( email ):
            if escaped:
                return msnp13.escape( self.contacts[ email ].getNick() )
            else:
                return self.contacts[ email ].getNick()
        else:
            raise ContactNotInListError, email

    def setContactNick( self, email, value ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setNick( value )
        else:
            raise ContactNotInListError, email

    def getContactPersonalMessage( self, email, escaped=False ):
        if self.contacts.has_key( email ):
            if escaped:
                return msnp13.escape( self.contacts[ email ].getPersonalMessage() )
            else:
                return self.contacts[ email ].getPersonalMessage()
        else:
            raise ContactNotInListError, email

    def setContactPersonalMessage( self, email, value ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setPersonalMessage( value )
        else:
            raise ContactNotInListError, email

    def getContactAlias( self, email, escaped=False ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getAlias()
        else:
            raise ContactNotInListError, email

    def setContactAlias( self, email, value ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setAlias( value )
        else:
            raise ContactNotInListError, email
            
    def getContactNameToDisplay( self, email ):
        displayName = self.getContactAlias( email, True )        
        if displayName == '':
            displayName = self.getContactNick( email, True )
            
        return displayName

    def getContactId( self, email ):
        if self.contacts.has_key( email ):
            return self.contacts[ email ].getId()
        else:
            raise ContactNotInListError, email

    def getContactGroupIds( self, email ):
        '''return a list with the group ids or an empty list'''

        if self.contacts.has_key( email ) and len( self.contacts[ email ].getGroups() ) > 0:
            return self.contacts[ email ].getGroups()
        elif self.contacts.has_key( email ):
            return []
        else:
            raise ContactNotInListError, email

    def getNoGroup( self ):
        return self.noGroup

    def unblockContact( self, email ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setBlocked( False )
            self.contacts[ email ].setAllow( True )
        else:
            raise ContactNotInListError, email

    def blockContact( self, email ):
        if self.contacts.has_key( email ):
            self.contacts[ email ].setBlocked( True )
            self.contacts[ email ].setAllow( False )
        else:
            raise ContactNotInListError, email

    def removeUserFromGroup( self, user, group ):
        if self.groups.has_key( group ):
            contact = self.groups[ group ].getUser( user )
            if contact != None:
                # remove group from user's list of belongings
                contact.removeGroup( group )

                # add the contact to no group if applicable
                if contact.getGroups() == []:
                    self.noGroup.setUser( user, contact )

                # remove user from group
                self.groups[ group ].removeUser( user )
            else:
                raise ContactNotInGroupError, ( user, group )
        else:
            raise GroupNotFoundError, group

    def removeContact( self, contactMail ):
        if self.contacts.has_key( contactMail ):
            # remove user from groups to which he belongs
            contact = self.contacts[ contactMail ]
            for group in contact.getGroups():
                self.groups[ group ].removeUser( contactMail )
            
            # remove user from the no group, if applicable
            if len( contact.getGroups() ) == 0:
                self.noGroup.removeUser( contactMail )
            
            # remove user
            del self.contacts[ contactMail ]
        else:
            raise ContactNotInListError, contactMail

    def addUserToGroup( self, user, group ):
        if self.groups.has_key( group ) and self.contacts.has_key( user ):
            contact = self.contacts[ user ]
            # remove from nogroup if applicable
            if len( contact.getGroups() ) == 0:
                self.noGroup.removeUser( user )
            
            groupObj = self.groups[ group ]
            contact.addGroup( group )
            
            groupObj.setUser( user, contact )
        elif not self.contacts.has_key( user ):
            raise ContactNotInListError, user
        elif group not in self.groups:
            raise GroupNotFoundError, group


    def setDynamicItemsXml( self, xml ):
        '''modify the structure with a new DynamicItems xml'''

        self.groups = {}
        self.contacts = {}

        doc = minidom.parseString( xml )
        
        # Retrieve groups
        for i in doc.getElementsByTagName( 'Group' ):
            groupId = i.getElementsByTagName( 'groupId' )[ 0 ].childNodes[ 0 ].data
            name = i.getElementsByTagName( 'name' )[ 0 ].childNodes[ 0 ].data

            if groupId not in self.groups:
                self.setGroup( groupId , Group( name, groupId ) )
        
        # Retrieve contacts
        for i in doc.getElementsByTagName( 'Contact' ):
            if i.getElementsByTagName( 'isMessengerUser' )[ 0 ].\
                                            childNodes[ 0 ].data == 'true':
                try:
                    email = i.getElementsByTagName( 'passportName' )[ 0 ].\
                                            childNodes[ 0 ].data
                except:
                    continue
            else:
                continue

            contact = Contact( email )

            contactId = i.getElementsByTagName( 'contactId' )[ 0 ].\
                                                childNodes[ 0 ].data
            contact.setId( contactId )

            groups = []

            for j in i.getElementsByTagName( 'guid' ):
                guid = j.childNodes[ 0 ].data
                groups.append( guid )
            contact.setGroups( groups )

            for j in i.getElementsByTagName( 'Annotation' ):
                try:
                    if j.getElementsByTagName( 'Name' )[ 0 ].\
                                        childNodes[ 0 ].data == 'AB.NickName':
                        alias = j.getElementsByTagName( 'Value' )[ 0 ].\
                                                        childNodes[ 0 ].data
                        contact.setAlias( alias )
                except:
                    pass

            displayName = i.getElementsByTagName( 'displayName' )[ 0 ].childNodes[ 0 ].data
            contact.setNick( displayName )

            isMobileIMEnabled = i.getElementsByTagName( 'isMobileIMEnabled' )[ 0 ]\
                                                        .childNodes[ 0 ].data
            contact.setMobile( isMobileIMEnabled == 'true' )

            hasSpace = i.getElementsByTagName( 'hasSpace' )[ 0 ].\
                                                    childNodes[ 0 ].data
            contact.setSpace( hasSpace == 'true' )
            
            # finally adds the contact
            self.addContact( contact )
            
        self.updateMemberships()

    def setMembershipListXml( self, xml ):
        '''modify the structure with a new MembershipList xml
        if it is the first xml that you send, send first the dinamyc items'''
        
        self.lists[ 'Allow' ] = []
        self.lists[ 'Block' ] = []
        self.lists[ 'Reverse' ] = []
        self.lists[ 'Pending' ] = []
        
        ml = minidom.parseString( xml )

        for i in ml.getElementsByTagName( 'Membership' ):
            memberRole = i.getElementsByTagName( 'MemberRole' )[ 0 ].childNodes[ 0 ].data

            for j in i.getElementsByTagName( 'Member' ):
                try:
                    email = j.getElementsByTagName( 'PassportName' )[ 0 ].childNodes[ 0 ].data

                    if email not in self.lists[ memberRole ]:
                        self.lists[ memberRole ].append( email )
                    else:
                        ec.debugmsg( str(email) + 'appeared twice in list ' \
                                     + str(memberRole) )
                except Exception, e:
                    pass
        
                
    def updateMemberships( self ):
        '''
        Updates contact membership info according to self.lists
        '''
        
        for email in self.contacts:
            if email in self.lists[ 'Reverse' ]:
                self.contacts[ email ].setReverse( True )
            else:
                self.contacts[ email ].setReverse( False )
                
            if email in self.lists[ 'Allow' ]:
                self.contacts[ email ].setAllow( True )
            else:
                self.contacts[ email ].setAllow( False )
                
            if email in self.lists[ 'Block' ]:
                self.contacts[ email ].setBlocked( True )
            else:
                self.contacts[ email ].setBlocked( False )

    def getADL( self ):
        '''return a list of XML for the ADL command, is a list because each ADL should be less than 7500
           bytes'''

        domains = {}

        for i in self.contacts.keys():
            (user, domain) = i.split( '@' )

            if domains.has_key( domain ):
                domains[ domain ].append( user )
            else:
                domains[ domain ] = [ user ]

        xmlDomains = []

        for i in domains.keys():
            users = ''
            for j in domains[ i ]:

                l = 0

                if self.getContact( j + '@' + i ).getAllow():
                    l = 3

                if self.getContact( j + '@' + i ).getBlocked():
                    l = 5

                if l > 0:
                    users += '<c n="' + j + '" l="' + str( l ) + '" t="1" />'

                if len( users ) + len( '<d n="' + i + '"></d>' ) > 7200:
                    xmlDomains.append( '<d n="' + i + '">' + users + '</d>' )
                    users = ''

            if len( users ) > 0:
                xmlDomains.append( '<d n="' + i + '">' + users + '</d>' )


        adls = []
        full = False

        while len( xmlDomains ) > 0:
            xml = '<ml l="1">'

            for i in range( len( xmlDomains ) ):
                domain = xmlDomains.pop()

                if len( xml ) + len( domain ) < 7400:
                    xml += domain
                else:
                    xml += '</ml>'
                    adls.append( xml )
                    xmlDomains.append( domain )
                    full = True
                    break

            if not full:
                xml += '</ml>'
                adls.append( xml )
            else:
                full = False
        
        return adls

    def getOnlineUsers( self ):
        '''
        return a list of online users
        '''

        lista = []
        for i in self.getContacts().keys():
            if self.getContactStatus( i ) != 'FLN':
                lista.append( [ i, self.getContactStatus( i ) ] )

        return lista
    
    def getOnlineUsersDict( self ):
        dictionary = {}
        
        for i in self.getContacts().keys():
            if self.getContactStatus( i ) != 'FLN':
                dictionary[ i ] = self.getContact( i )

        return dictionary

    def getOnOffUsersRelationByGroup(self, groupName):
        #return a 2 tuple containing the relation of users online and offline
        groupSizeStr = ''
        groupObject = None

        if groupName == 'No group':
            groupObject = self.getNoGroup()
        else:
            try:
                groupObject = self.reverseGroups[ groupName ]
            except KeyError:
                raise GroupNotFoundError, groupName

        if groupObject != None:
            groupSize = groupObject.getSize()
            usersOnline = groupObject.getOnlineUsersNumber()

        return usersOnline, groupSize

    def getContactList( self, showOffline = True, showEmptyGroups = False, \
            orderByStatus = False ):
        '''
        return a dictionarie with the contact list sorted acording the parameters
        '''

        cl = {}

        if orderByStatus:
            cl[ 'offline' ] = {}
            cl[ 'online' ] = {}
            for email in self.contacts:
                status = self.getContactStatus( email )
                if status == 'FLN' and not showOffline:
                    continue
                elif status == 'FLN':
                    cl[ 'offline' ][ email ] = self.contacts[ email ]
                else:
                    cl[ 'online' ][ email ] = self.contacts[ email ]
        else:
            # Initialize return dict and build id2group_name dict
            for i in self.reverseGroups:
                cl[ i ] = {}
            cl[ 'No group' ] = {}
                
            # classify contacts into their group/s
            for email in self.contacts:
                contactGroups = []
                for id in self.getContactGroupIds( email ):
                    contactGroups += [ self.groups[ id ].getName() ]    
                if len( contactGroups ) == 0: # email doesn't belong to any group
                    contactGroups = [ 'No group' ]
                # the actual classification:
                for group in contactGroups:
                    if showOffline:
                            cl[ group ][ email ] = self.contacts[ email ]
                    elif self.getContactStatus( email ) != 'FLN':
                            cl[ group ][ email ] = self.contacts[ email ]

        if not showEmptyGroups:
            for i in cl.keys():
                if len( cl[ i ] ) == 0:
                    del cl[ i ]

        return cl

def showCl( cl ):
    for (key, value) in cl.getGroups().iteritems():
        print key + ' ' + value.getId()
        def show( email, contact ):
            print '\t' + email + ' ' + contact.getNick() + ' ',

            if contact.getAlias() != '':
                print contact.getAlias() + ' ',

            if contact.getMobile():
                print 'Mobile ',

            if contact.getAllow():
                print 'Allowed ',

            if contact.getReverse():
                print 'Reverse ',

            if contact.getPending():
                print 'Pending ',

            if contact.getBlocked():
                print 'Blocked ',

            if contact.getSpace():
                print 'Space ',

            print contact.getGroups()

        for (email, contact) in value.getUsers().iteritems():
            show( email, contact )

    print 'No Goup'
    for (email, contact) in cl.noGroup.getUsers().iteritems():
        show( email, contact )
        
class ContactNotInListError( Exception ):
    def __init__( self, value ):
        self.value = value
        
    def __str__( self ):
        return 'Contact ' + repr( self.value ) + ' is not in the list'
    
class ContactNotInGroupError( Exception ):
    def __init__( self, value, group ):
        self.value = value
        self.group = group
        
    def __str__( self ):
        return 'Contact ' + repr( self.value ) + ' is not in this group: ' \
                        + str( self.group )
                        
class GroupNotFoundError( Exception ):
    def __init__( self, value ):
        self.value = value
        
    def __str__( self ):
        return 'Group ' + repr( self.value ) + ' does not exist'
