# -*- coding: utf-8 -*-

#   This file is part of emesene.
#
#    Emesene is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    emesene is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with emesene; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import gtk
import gobject
import Dialogs
import emesenecommon as ec
from emesenecommon import debugmsg

class Plugin( object ):
    '''
    Base class to make a plugin
    '''
    
    def __init__( self, controller, msn, period = None ):
        '''
        Contructor, the period parameter is the time between calls to the function, in milliseconds,
        if period = None then the plugin is asynchronous.
        '''
        
        self.controller = controller
        self.msn = msn
        
        self.period = period
        
        self.name = ''
        self.description = ''
        # name : mail
        self.authors = {}
        self.website = ''
        self.enabled = False
        
    def check( self ):
        '''
        check if everything is OK to start the plugin
        return a tuple whith a boolean and a message
        if OK -> ( True , 'some message' )
        else -> ( False , 'error message' )
        '''
        
        pass
    
    def start( self ):
        '''
        start the plugin
        '''
        
        pass
    
    def stop( self ):
        '''
        stop the plugin
        '''
        
        pass
    
    def action( self ):
        '''
        the method that is called every self.period milliseconds
        '''
        
        pass
    
    def connect( self, signal, callback ):
        '''
        connect a signal to a callback
        '''
        
        self.msn.connect( signal, callback )
        
    def configure( self ):
        '''display a configuration dialog'''
        
        Dialogs.Information( None, _( 'The plugin has no preferences' ) ).run()
        
    def disconnect( self, signal, callback ):
        '''
        disconnect a signal to a callback
        '''
        
        self.msn.disconnect( signal, callback )
        
    def isEnabled( self ):
        '''
        return True if the plugin is running
        '''
        
        return self.enabled
    
class Option( object ):
    '''a type of option to use by the plugins'''
    
    def __init__( self, name, optionType, label, description, value, options = None ):
        '''contructor
        name: represent the name of the variable
        optionType: the type expected (str, int, float, bool, list)
        label: the text that will appear in the label
        descrption: the text in the tooltip
        value: the default value when show and the input value when returned'''
        
        self.name = name
        self.optionType = optionType
        self.label = label
        self.description = description
        self.value = value
        self.options = options

class ConfigWindow( gtk.Dialog ):
    '''build a generic window with configuration parameters to fill'''
    
    def __init__( self, title, optionList ):
        '''contructor'''
        
        gtk.Window.__init__( self )
        self.set_title( title )
        self.set_border_width( 6 )
        
        vbox = gtk.VBox()
        vbox.set_spacing( 12 )
        
        self.add_button(gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL)
        self.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        
        self.options = {}
        self.widgets = {}
        
        for i in optionList:
            self.options[ i.name ] = i

            hbox = gtk.HBox()
            if i.optionType == bool:
                check = gtk.CheckButton( i.label )

                if type( i.value ) == bool:
                    check.set_active( i.value )
                
                self.widgets[ i.name ] = check.get_active
                hbox.pack_start( check )
            elif i.optionType == list:
                if type( i.options ) == list:
                    label = gtk.Label( i.label )
                    label.set_justify(gtk.JUSTIFY_LEFT)
                    label.set_width_chars( 20 )
                    
                    combo = gtk.combo_box_entry_new_text()
                    self.widgets[ i.name ] = combo.get_active_text
                    hbox.pack_start( label )
                    hbox.pack_start( combo )
                    
                    count = 0
                    done = False
                    for j in i.options:
                        if not done:
                            if j == i.value:
                                done = True
                            else:
                                count += 1
                            
                        combo.append_text( str( j ) )
                    
                    if not done:
                        combo.set_active( 0 )
                    else:   
                        combo.set_active( count )
                else:
                    def dummy():
                        return None
                    
                    self.widgets[ i.name ] = dummy
            
            else:
                label = gtk.Label( i.label )
                label.set_justify(gtk.JUSTIFY_LEFT)
                label.set_width_chars( 20 )
                
                entry = gtk.Entry()
                entry.set_text( str( i.value ) )
            
                self.widgets[ i.name ] = entry.get_text
                hbox.pack_start( label )
                hbox.pack_start( entry )
                
            vbox.pack_start( hbox )
        
        vbox.show_all()
        
        self.vbox.pack_start( vbox )
        
        self.connect( 'delete-event', self.close )
        
    def run( self, *args ):
        response = gtk.Dialog.run( self )
        self.hide()
        
        if response == gtk.RESPONSE_OK:
            for i in self.options.keys():
                self.options[ i ].value = self.widgets[ i ]()
                
            return self.options
        else:
            return None

    def close( self, *args ):
        self.response( gtk.RESPONSE_CLOSE )

class Config:
    '''This class contains all the stuff relative to the configuration of emesene.
    ou can get or set configuration values and will be saved to a file for persistence'''

    def __init__( self, pluginName ):
        '''Constructor'''
        
        self.pluginName = pluginName
        self.name = ec.PLUGIN_HOME_PATH + ec.DIR_SEP + pluginName + '.conf'
        self.configDict = {}
        self.readConfig()
        
    def readConfig( self ):
        '''read the config file and create a dictionarie with key and value
        of all the key=value\n in the config file'''
        
        conf = None
        try:
            conf = file( self.name , "r" )
            
            string = conf.read()
            string = string.replace( " " , "" )
    
            lines = string.split( "\n" )

            for i in lines:
                if i == "":
                    continue
                
                try:
                    key,value = i.split( "=" )

                    self.configDict[ key ] = value.replace( '%20', ' ' )
                except Exception, e:
                    debugmsg( key + ' plugin config value is incorrect' )
                    
            conf.close()
        except Exception, e:
            if conf:
                conf.close()
#            debugmsg( 'Coudn\'t read plugin\'s conf file:\n\t' + self.name )
#            debugmsg( 'Reason: ' + str( e ) )
            
    def writeConfig( self ):
        '''write the config to the file, overwrite current config file'''
        
        conf = None
        try:
            conf = file( self.name, "w" )
        
            for i in self.configDict.keys():
                conf.write( i + "=" + self.configDict[ i ] + "\n" )

            conf.flush()
            conf.close()
        except Exception, e:
            conf.close()
            debugmsg( 'Coudn\'t write plugin\'s conf file:\n\t' + self.name )
            debugmsg( 'Reason: ' + str( e ) )
        
    def getValue( self , key, default='' ):
        '''get the value from key in configDict, return default if not found'''
        
        try:
            return self.configDict[ key ].replace( '%20', ' ' )
        except:
            debugmsg( self.pluginName + ': config value not found ' + key )
            self.setValue(key, default)
            self.writeConfig()
            return default

    def setValue( self , key , value ):
        '''set a value in configDict, to make it last to the next session
        you have to call writeConfig before you close'''
        
        self.configDict[ key ] = str( value ).replace( ' ', '%20' )
        self.writeConfig()
